package edu.uci.ics.cmi.HistorySlicingFuzzy.parser;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import edu.uci.ics.cmi.CodeRepository.model.Diff;
import edu.uci.ics.cmi.CodeRepository.model.Operation;

public class DiffProcessorFuzzy {
	
	public DiffProcessorFuzzy() {
	}

	public List<Diff> getChangeBlockDiffs(String[][] changeText, int leftFrom, int leftTo, int rightFrom, int rightTo, Integer commitId) {
		List<Diff> diffs = new LinkedList<Diff>();
				
		int [] from = {leftFrom, rightFrom};
		int [] to = {leftTo, rightTo};
		
		HashMap<Long, Diff> leftToRight = getDiffsWithFuzziness(commitId, changeText, from, to, false);
		HashMap<Long, Diff> rightToLeft = getDiffsWithFuzziness(commitId, changeText, from, to, true);

		// rightToLeft lists
		for (Long key: leftToRight.keySet())
		{
			Diff dS = leftToRight.get(key);
			
			if (!rightToLeft.containsKey(key))
			{
				rightToLeft.put(key, dS);
			}
			else
			{
				Diff dM = rightToLeft.get(key);
				if (dS.getWeight() < dM.getWeight())
				{
					rightToLeft.remove(key);
					rightToLeft.put(key, dS);
				}
			}
			
		}
		
		diffs.addAll(rightToLeft.values());
		
		diffs = filterDiffs(diffs, threshold, leftFrom, leftTo, rightFrom, rightTo, commitId);
		
		return diffs;
	}

	private HashMap<Long, Diff> getDiffsWithFuzziness(Integer commitId, String[][] changeText, int [] from, int [] to, boolean backwards) {
		
		HashMap<Long, Diff> diffs = new HashMap<Long, Diff>();

		// Configure order of traversal of the diff
		int left = 0; 
		int right = 1;
		if (backwards)
		{
			left = 1;
			right = 0;
		}
		
		int numberOfLinesLeft = to[left] - from[left] + 1;
		int numberOfLinesRight = to[right] - from[right] + 1;

		// For every line on the left
		for (int i = 0; i < numberOfLinesLeft; i++)
		{
			String leftString = changeText[left][i];
			
			Float [] mapping = getMinimumDistanceMapping (leftString, changeText[right], numberOfLinesRight);
			HashMap<Long, Diff> foundDiffs = getDiffs(commitId, from[left] + i, from[right], mapping, backwards);
			for (Diff d: foundDiffs.values())
			{
				if (diffs.containsKey(d.hashKeyFromTo()))
					System.out.println("collision!");
				diffs.put(d.hashKeyFromTo(), d);
			}
		}

		return diffs;
	}
	
	private HashMap<Long, Diff> getDiffs(Integer commitId, Integer from, Integer to, Float [] differences, boolean backwards) {
		HashMap<Long, Diff> diffs = new HashMap<Long, Diff>();
		
		for (int i = 0; i < differences.length; i++)
		{
			Diff diff;
			if (!backwards)
				diff = new Diff(commitId, from, Operation.CHANGE, to + i);
			else
				diff = new Diff(commitId, to + i, Operation.CHANGE, from);

			diff.setWeight(differences[i]);
			diffs.put(diff.hashKeyFromTo(), diff);
		}
				
		return diffs;
	}
	
	public Float [] getMinimumDistanceMapping (String left, String [] right, int maxNumberOfLines)
	{
		// Save the candidates that were found at any iteration of depth
		HashMap<CandidateId, Float> candidatesFound = new HashMap<CandidateId, Float>();

		// CandidateIds is an array of bit arrays representing the candidate combinations of lines at each iteration of depth
		ArrayList<CandidateId> candidateIds = new ArrayList<CandidateId>();

		// CandidateStrings is an array of strings representing the candidate combinations of lines at each iteration of depth
		// This array is necessary to preserve the actual order in which the lines were combined
		ArrayList<String> candidateStrings = new ArrayList<String>();
		
		// Initialize candidates to just each individual line (without combined lines)
		for (int i = 0; i < maxNumberOfLines; i++)
		{
			CandidateId candidateId = new CandidateId(maxNumberOfLines);
			candidateId.add(i);
			String candidateString = right[i];
			Float candidateDistance = compareLines(left, candidateString);

			candidateIds.add(candidateId);
			candidateStrings.add(candidateString);
			candidatesFound.put(candidateId, candidateDistance);
		}
		
		// For each depth (number of additional lines to join)
		int maxDepth = 2;
		for (int depth = 0; depth < maxDepth; depth++)
		{
			// Initialize list of candidates for next iteration
			ArrayList<CandidateId> newCandidateIds = new ArrayList<CandidateId>();
			ArrayList<String> newCandidateStrings = new ArrayList<String>();

			// Consider each candidate (a candidate is a combination of lines)
			for (int c = 0; c < candidateIds.size(); c++)
			{
				// Get characteristics of the candidate
				CandidateId candidateId = candidateIds.get(c);
				String candidateString = candidateStrings.get(c);
				Float candidateDistance = candidatesFound.get(candidateId);
				
				// for any other line that is not included in the candidate
				for (int nextLine = 0; nextLine < maxNumberOfLines; nextLine++)
				{
//					if (!lineInLineId(nextLine, candidateId))
					if (!candidateId.containsLine(nextLine))
					{
						// Get characteristics of the line to be combined
						CandidateId nextLineId = new CandidateId(maxNumberOfLines);
						nextLineId.add(nextLine);
//						CandidateId nextLineId = getLineId(nextLine, maxNumberOfLines);
						String nextLineString = right[nextLine];
						Float nextLineDistance = candidatesFound.get(nextLineId);

						// Get combined candidate (the combined candidate is only saved in candidateFounds 
						// if it has a shorter distance that both of its components)
						CandidateId combinedId = candidateId.getCombinedId(nextLineId);
						String combinedString = candidateString + nextLineString;
						Float combinedDistance = compareLines(left, combinedString);

						// If the combined candidate has a shorter distance than both of its components,
						// save it as a "good" one, and keep iterating on it
						if ( (combinedDistance < candidateDistance) && (combinedDistance < nextLineDistance) )
						{
							boolean saveCandidate = true;
							// This combination of lines may have been already saved in a different order
							// Only save this combination if it is better than the older combination 
							if (candidatesFound.containsKey(combinedId))
								if (combinedDistance < candidatesFound.get(combinedId))
									candidatesFound.remove(combinedId);
								else
									saveCandidate = false;
							
							if (saveCandidate)
							{
								// Save the combined candidate for the next iteration
								newCandidateIds.add(combinedId);
								newCandidateStrings.add(combinedString);

								// Save the candidate as a "good" candidate
								candidatesFound.put(combinedId, combinedDistance);
							}
						}
					}
				}
			}
			
			// Initialize next iteration
			candidateIds = newCandidateIds;
			candidateStrings = newCandidateStrings;
		}
		
		Float [] minimumDistances = new Float[maxNumberOfLines];
		for (int i = 0; i < minimumDistances.length; i++)
			minimumDistances[i] = Float.MAX_VALUE;
		
		// for each line, return the minimum distance in any combination
		for (CandidateId candidateId: candidatesFound.keySet())
		{
			Float candidateDistance = candidatesFound.get(candidateId);
			
			for (int line = 0; line < maxNumberOfLines; line++)
				if (candidateId.containsLine(line))
					if (candidateDistance < minimumDistances[line])
						minimumDistances[line] = candidateDistance;
		}
		
		return minimumDistances;
	}

}
