package edu.uci.ics.cmi.HistorySlicingFuzzy;

public class EvaluateHistorySlices {

	public static void main(String[] args) throws IOException, ParseException, SQLException, InstantiationException, IllegalAccessException, ClassNotFoundException {

		if (args.length < 1)
		{
			System.out.println("Please, specify a configuration file.");
			return;
		}

		// Read configuration from file
		configurationFile = args[0];

		evaluate();
	}

	private static String configurationFile;

	/**
	 * @throws IOException
	 * @throws ParseException
	 * @throws SQLException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws ClassNotFoundException
	 */
	protected static void evaluate() throws IOException, ParseException, SQLException, InstantiationException, IllegalAccessException, ClassNotFoundException
	{
		Configuration.load(configurationFile);

		Subject subject = new Subject(configurationFile);
		Settings settings = new Settings(configurationFile);
		HistorySlicingDBFuzzy db = new HistorySlicingDBFuzzy(configurationFile);
		FileAccess fileAccess = new FileAccess(configurationFile);

		List<String> historySlicingSeeds = Configuration.getPropertyListString("cmi.historySlicing.evaluate.historySlicingSeeds");
		List<List<Integer>> historySlicingSeedsGroundTruth = Configuration.getPropertyListsInteger("cmi.historySlicing.evaluate.historySlicingSeeds.groundTruth");
		List<Integer> historySlicingSeedsMethodSize = Configuration.getPropertyListInteger("cmi.historySlicing.evaluate.historySlicingSeeds.methodSize");
		List<Integer> historySlicingSeedsRevisionBackToCheck = Configuration.getPropertyListInteger("cmi.historySlicing.evaluate.historySlicingSeeds.revisionBackToCheck");

		CodeRepository codeRepository = CodeRepository.createCodeRepository(configurationFile);
		if (codeRepository == null)
			return;

		CodeRepositoryParser codeRepositoryParser = CodeRepositoryParser.createCodeRepositoryParser(configurationFile);
		if (codeRepositoryParser == null)
			return;

		String technique;
		if (db.getUrl().endsWith("Fuzzy"))
			technique = "fuzzy";
		else
			technique = "1to1";
		
		Integer ndcgLevel = 50;
		System.out.print("\"project\",\"technique\",\"historySlicingSeedId\",\"positions found\",\"effectiveness\"\n");
		// For each seed
		for (int i = 0; i < historySlicingSeeds.size(); i++)
		{
			String seedFile = historySlicingSeeds.get(i);
			List<Integer> groundTruth = historySlicingSeedsGroundTruth.get(i);
			Integer methodSize = historySlicingSeedsMethodSize.get(i);
			Integer revisionBackToCheck = historySlicingSeedsRevisionBackToCheck.get(i);
			
			HistorySlicingSeed seed = (HistorySlicingSeed) fileAccess.readXmlObject(seedFile);

			HistorySlicerFuzzy slicer = new HistorySlicerFuzzy(db, subject, settings, codeRepository, codeRepositoryParser, seed);
			slicer.executeFuzzy();

			HistorySliceFuzzy slice = slicer.getSliceFuzzy();
			
			LinkedList<Integer> positionsFound = getPositionsFound(slice, groundTruth, revisionBackToCheck, true);
			float goodness = NDCG(positionsFound, groundTruth, ndcgLevel);
//			double goodness = computeEffectiveness(positionsFound, groundTruth, methodSize, methodSize);
			
			System.out.print("\"");
			System.out.print(subject.getProject());
			System.out.print("\",\"");
			System.out.print(technique);
			System.out.print("\",");
			System.out.print(i + 1);
			System.out.print(",\"");
			System.out.print(positionsFound);
			System.out.print("\",");
			System.out.print(goodness);
			System.out.print("\n");
			
			if (technique.contentEquals("fuzzy"))
			{
				positionsFound = getPositionsFound(slice, groundTruth, revisionBackToCheck, false);
				goodness = NDCG(positionsFound, groundTruth, ndcgLevel);
				
				System.out.print("\"");
				System.out.print(subject.getProject());
				System.out.print("\",\"");
				System.out.print("annotation");
				System.out.print("\",");
				System.out.print(i + 1);
				System.out.print(",\"");
				System.out.print(positionsFound);
				System.out.print("\",");
				System.out.print(goodness);
				System.out.print("\n");
			}
			
		}
		
	}

	
	
	
	

	public static TreeMap<Float, LinkedList<Integer>> getRecommendedLines(HistorySliceFuzzy slice, Integer revisionBackToCheck, boolean accountForWeights)
	{
		// Get the scores for the slice nodes
		HashMap<Integer, Float> previousNodeScores = new HashMap<Integer, Float>();

		// For every revision back to check
		for (int j = 0; j < (revisionBackToCheck - 1); j++)
		{
			HashMap<Integer, Float> currentNodeScores = new HashMap<Integer, Float>();

			// Iterate through this revision's edges
			for (int k = 0; k < slice.getChangesReverseFuzzySize(0, j); k++)
			{
				Diff d = slice.getChangeReverseFuzzy(0, j, k);

				float pathScore = 1f;
				
				if (accountForWeights)
					pathScore -= d.getWeight();

				// If this edge is connected (it has a longer path), weight its score 
				if (previousNodeScores.containsKey(d.getNewLine()))
					pathScore = pathScore * previousNodeScores.get(d.getNewLine());

				float currentNodeScore = 0f;

				// If this is not the first path reaching this node
				if (currentNodeScores.containsKey(d.getOldLine()))
				{
					currentNodeScore = currentNodeScores.get(d.getOldLine());
					currentNodeScores.remove(d.getOldLine());
				}

				currentNodeScores.put(d.getOldLine(), Math.max(currentNodeScore, pathScore));

			}

			// Prepare for next iteration
			previousNodeScores = currentNodeScores;
		}

		// Sort the output
		TreeMap<Float, LinkedList<Integer>> sortedLines = new TreeMap<Float, LinkedList<Integer>>(Collections.reverseOrder());
		for (Integer node: previousNodeScores.keySet())
		{
			float score = previousNodeScores.get(node);

			if (!sortedLines.containsKey(score))
				sortedLines.put(score, new LinkedList<Integer>());

			sortedLines.get(score).add(node);
		}

		return sortedLines;
	}

	
	public static LinkedList<Integer> getPositionsFound(HistorySliceFuzzy slice, List<Integer> groundTruth, Integer revisionBackToCheck, boolean accountForWeights)
	{
		LinkedList<Integer> positionsFound = new LinkedList<Integer>();

		if (slice.getNChangesFuzzy(0) < revisionBackToCheck)
			return positionsFound;
		
		TreeMap<Float, LinkedList<Integer>> recommendedLines = getRecommendedLines(slice, revisionBackToCheck, accountForWeights);

		// Save positions of lines that were found in the recommended list
		Integer currentPosition = 0; 
		for (Float score: recommendedLines.keySet())
		{
			LinkedList<Integer> lines = recommendedLines.get(score);

			Integer nFound = 0;
			for (Integer line: lines)
				if (groundTruth.contains(line))
					nFound++;

			LinkedList<Integer> averagePositions = averagePositions(nFound, lines.size());
			for (Integer apos: averagePositions)
				positionsFound.add(currentPosition + apos); 
			
			currentPosition += lines.size();
		}
		
		return positionsFound;
	}
	
	public static LinkedList<Integer> averagePositions (Integer nFound, Integer nTied)
	{
		LinkedList<Integer> average = new LinkedList<Integer>();
		
		Integer wrong = nTied - nFound;
		
		Integer halfWrong = (int) Math.ceil(((double) (wrong)) / 2);
		
		for (int i = 0; i < nFound; i++)
			average.add(halfWrong + i);
		
		return average;
	}
	
	public static float NDCG(LinkedList<Integer> positionsFound, List<Integer> groundTruth, Integer ndcgLevel)
	{
		float relevanceScore = 1f;
		
		float dcg = 0f;
		for (Integer pos: positionsFound)
			if (pos < ndcgLevel)
				dcg += (Math.pow(2, relevanceScore) - 1) * Math.log(2) / Math.log(pos + 2);
		
		float idcg = 0f;
		for (int pos = 0; pos < groundTruth.size(); pos++)
			if (pos < ndcgLevel)
				idcg += (Math.pow(2, relevanceScore) - 1) * Math.log(2) / Math.log(pos + 2);
		
		float ndcg = dcg / idcg;
		
		return ndcg;
	}
	
	public static float computeEffectiveness(LinkedList<Float> positionsFound, List<Integer> groundTruth, Integer randomRecommendationSize, Integer methodSize)
	{
		if (randomRecommendationSize > methodSize)
			randomRecommendationSize = methodSize;

		// Calculate average case of the position of lines that were NOT found in the recommended list
		int foundLines = positionsFound.size();
		for (int i = foundLines; i < groundTruth.size(); i++)
		{
			positionsFound.add(foundLines + ((float) (randomRecommendationSize - 1 - foundLines)) / 2);
		}
		
		// Average position in which the ground truth lines were recommended
		float averagePosition = 0f;
		for (Float pos: positionsFound)
			averagePosition += pos;
		averagePosition = averagePosition / positionsFound.size();
		
		// Normalized score
		float score = 1f - averagePosition / (methodSize - 1);
		return score;	
	}


}
